@@ -13,8 +13,8 @@ use tauri_runtime::{
     dpi::{LogicalPosition, LogicalSize, PhysicalPosition, PhysicalSize, Position, Size},
     DetachedWindow, PendingWindow, WindowEvent,
   },
-  ClipboardManager, Dispatch, Error, GlobalShortcutManager, Icon, Params, Result, RunEvent,
-  RunIteration, Runtime, RuntimeHandle, UserAttentionType,
+  ClipboardManager, Dispatch, Error, GlobalShortcutManager, Icon, Result, RunEvent, RunIteration,
+  Runtime, RuntimeHandle, UserAttentionType,
 };
 
 #[cfg(feature = "menu")]
@@ -23,8 +23,11 @@ use tauri_runtime::window::MenuEvent;
 use tauri_runtime::{SystemTray, SystemTrayEvent};
 #[cfg(windows)]
 use winapi::shared::windef::HWND;
+#[cfg(target_os = "linux")]
+use wry::application::platform::unix::{WindowBuilderExtUnix, WindowExtUnix};
 #[cfg(windows)]
-use wry::application::platform::windows::WindowBuilderExtWindows;
+use wry::application::platform::windows::{WindowBuilderExtWindows, WindowExtWindows};
+
 #[cfg(feature = "system-tray")]
 use wry::application::system_tray::{SystemTray as WrySystemTray, SystemTrayBuilder};
 
@@ -74,9 +77,6 @@ use menu::*;
 mod mime_type;
 use mime_type::MimeType;
 
-type MainTask = Arc<Mutex<Option<Box<dyn FnOnce() + Send>>>>;
-type CreateWebviewHandler =
-  Box<dyn FnOnce(&EventLoopWindowTarget<Message>) -> Result<WebviewWrapper> + Send>;
 type WindowEventHandler = Box<dyn Fn(&WindowEvent) + Send>;
 type WindowEventListenersMap = Arc<Mutex<HashMap<Uuid, WindowEventHandler>>>;
 type WindowEventListeners = Arc<Mutex<HashMap<WindowId, WindowEventListenersMap>>>;
@@ -131,7 +131,7 @@ unsafe impl Send for GlobalShortcutWrapper {}
 #[derive(Clone)]
 pub struct GlobalShortcutManagerHandle {
   context: EventLoopContext,
-  shortcuts: HashMap<String, (AcceleratorId, GlobalShortcutWrapper)>,
+  shortcuts: Arc<Mutex<HashMap<String, (AcceleratorId, GlobalShortcutWrapper)>>>,
   listeners: GlobalShortcutListeners,
 }
 
@@ -159,7 +159,11 @@ impl GlobalShortcutManager for GlobalShortcutManagerHandle {
     )?;
 
     self.listeners.lock().unwrap().insert(id, Box::new(handler));
-    self.shortcuts.insert(accelerator.into(), (id, shortcut));
+    self
+      .shortcuts
+      .lock()
+      .unwrap()
+      .insert(accelerator.into(), (id, shortcut));
 
     Ok(())
   }
@@ -172,12 +176,12 @@ impl GlobalShortcutManager for GlobalShortcutManagerHandle {
       Message::GlobalShortcut(GlobalShortcutMessage::UnregisterAll(tx))
     )?;
     self.listeners.lock().unwrap().clear();
-    self.shortcuts.clear();
+    self.shortcuts.lock().unwrap().clear();
     Ok(())
   }
 
   fn unregister(&mut self, accelerator: &str) -> Result<()> {
-    if let Some((accelerator_id, shortcut)) = self.shortcuts.remove(accelerator) {
+    if let Some((accelerator_id, shortcut)) = self.shortcuts.lock().unwrap().remove(accelerator) {
       let (tx, rx) = channel();
       getter!(
         self,
@@ -406,7 +410,7 @@ pub struct WindowBuilderWrapper {
   inner: WryWindowBuilder,
   center: bool,
   #[cfg(feature = "menu")]
-  menu: Menu<u16>,
+  menu: Menu,
 }
 
 // safe since `menu_items` are read only here
@@ -441,15 +445,15 @@ impl WindowBuilder for WindowBuilderWrapper {
       window = window.position(x, y);
     }
 
-    if config.focus {
-      window = window.focus();
+    if config.center {
+      window = window.center();
     }
 
     window
   }
 
   #[cfg(feature = "menu")]
-  fn menu<I: MenuId>(mut self, menu: Menu<I>) -> Self {
+  fn menu(mut self, menu: Menu) -> Self {
     self.menu = convert_menu_id(Menu::new(), menu);
     self
   }
@@ -506,8 +510,9 @@ impl WindowBuilder for WindowBuilderWrapper {
     self
   }
 
-  fn focus(mut self) -> Self {
-    self.inner = self.inner.with_focus();
+  /// Deprecated since 0.1.4 (noop)
+  /// Windows is automatically focused when created.
+  fn focus(self) -> Self {
     self
   }
 
@@ -555,11 +560,17 @@ impl WindowBuilder for WindowBuilderWrapper {
     Ok(self)
   }
 
+  #[cfg(any(target_os = "windows", target_os = "linux"))]
   fn skip_taskbar(mut self, skip: bool) -> Self {
     self.inner = self.inner.with_skip_taskbar(skip);
     self
   }
 
+  #[cfg(target_os = "macos")]
+  fn skip_taskbar(self, _skip: bool) -> Self {
+    self
+  }
+
   fn has_icon(&self) -> bool {
     self.inner.window.window_icon.is_some()
   }
@@ -588,7 +599,9 @@ impl From<FileDropEventWrapper> for FileDropEvent {
     match event.0 {
       WryFileDropEvent::Hovered(paths) => FileDropEvent::Hovered(paths),
       WryFileDropEvent::Dropped(paths) => FileDropEvent::Dropped(paths),
-      WryFileDropEvent::Cancelled => FileDropEvent::Cancelled,
+      // default to cancelled
+      // FIXME(maybe): Add `FileDropEvent::Unknown` event?
+      _ => FileDropEvent::Cancelled,
     }
   }
 }
@@ -598,6 +611,23 @@ struct Hwnd(HWND);
 #[cfg(windows)]
 unsafe impl Send for Hwnd {}
 
+#[cfg(any(
+  target_os = "linux",
+  target_os = "dragonfly",
+  target_os = "freebsd",
+  target_os = "netbsd",
+  target_os = "openbsd"
+))]
+struct GtkWindow(gtk::ApplicationWindow);
+#[cfg(any(
+  target_os = "linux",
+  target_os = "dragonfly",
+  target_os = "freebsd",
+  target_os = "netbsd",
+  target_os = "openbsd"
+))]
+unsafe impl Send for GtkWindow {}
+
 #[derive(Debug, Clone)]
 enum WindowMessage {
   // Getters
@@ -618,6 +648,14 @@ enum WindowMessage {
   AvailableMonitors(Sender<Vec<MonitorHandle>>),
   #[cfg(windows)]
   Hwnd(Sender<Hwnd>),
+  #[cfg(any(
+    target_os = "linux",
+    target_os = "dragonfly",
+    target_os = "freebsd",
+    target_os = "netbsd",
+    target_os = "openbsd"
+  ))]
+  GtkWindow(Sender<GtkWindow>),
   // Setters
   Center(Sender<Result<()>>),
   RequestUserAttention(Option<UserAttentionTypeWrapper>),
@@ -678,14 +716,16 @@ pub(crate) enum ClipboardMessage {
   ReadText(Sender<Option<String>>),
 }
 
-#[derive(Clone)]
 pub(crate) enum Message {
-  Task(MainTask),
+  Task(Box<dyn FnOnce() + Send>),
   Window(WindowId, WindowMessage),
   Webview(WindowId, WebviewMessage),
   #[cfg(feature = "system-tray")]
   Tray(TrayMessage),
-  CreateWebview(Arc<Mutex<Option<CreateWebviewHandler>>>, Sender<WindowId>),
+  CreateWebview(
+    Box<dyn FnOnce(&EventLoopWindowTarget<Message>) -> Result<WebviewWrapper> + Send>,
+    Sender<WindowId>,
+  ),
   GlobalShortcut(GlobalShortcutMessage),
   Clipboard(ClipboardMessage),
 }
@@ -715,7 +755,7 @@ impl Dispatch for WryDispatcher {
     self
       .context
       .proxy
-      .send_event(Message::Task(Arc::new(Mutex::new(Some(Box::new(f))))))
+      .send_event(Message::Task(Box::new(f)))
       .map_err(|_| Error::FailedToSendMessage)
   }
 
@@ -827,6 +867,18 @@ impl Dispatch for WryDispatcher {
     Ok(dispatcher_getter!(self, WindowMessage::Hwnd).0)
   }
 
+  /// Returns the `ApplicatonWindow` from gtk crate that is used by this window.
+  #[cfg(any(
+    target_os = "linux",
+    target_os = "dragonfly",
+    target_os = "freebsd",
+    target_os = "netbsd",
+    target_os = "openbsd"
+  ))]
+  fn gtk_window(&self) -> Result<gtk::ApplicationWindow> {
+    Ok(dispatcher_getter!(self, WindowMessage::GtkWindow).0)
+  }
+
   // Setters
 
   fn center(&self) -> Result<()> {
@@ -854,10 +906,10 @@ impl Dispatch for WryDispatcher {
 
   // Creates a window by dispatching a message to the event loop.
   // Note that this must be called from a separate thread, otherwise the channel will introduce a deadlock.
-  fn create_window<P: Params<Runtime = Self::Runtime>>(
+  fn create_window(
     &mut self,
-    pending: PendingWindow<P>,
-  ) -> Result<DetachedWindow<P>> {
+    pending: PendingWindow<Self::Runtime>,
+  ) -> Result<DetachedWindow<Self::Runtime>> {
     let (tx, rx) = channel();
     let label = pending.label.clone();
     let context = self.context.clone();
@@ -866,9 +918,7 @@ impl Dispatch for WryDispatcher {
       .context
       .proxy
       .send_event(Message::CreateWebview(
-        Arc::new(Mutex::new(Some(Box::new(move |event_loop| {
-          create_webview(event_loop, context, pending)
-        })))),
+        Box::new(move |event_loop| create_webview(event_loop, context, pending)),
         tx,
       ))
       .map_err(|_| Error::FailedToSendMessage)?;
@@ -1161,10 +1211,10 @@ impl RuntimeHandle for WryHandle {
 
   // Creates a window by dispatching a message to the event loop.
   // Note that this must be called from a separate thread, otherwise the channel will introduce a deadlock.
-  fn create_window<P: Params<Runtime = Self::Runtime>>(
+  fn create_window(
     &self,
-    pending: PendingWindow<P>,
-  ) -> Result<DetachedWindow<P>> {
+    pending: PendingWindow<Self::Runtime>,
+  ) -> Result<DetachedWindow<Self::Runtime>> {
     let (tx, rx) = channel();
     let label = pending.label.clone();
     let dispatcher_context = self.dispatcher_context.clone();
@@ -1172,9 +1222,7 @@ impl RuntimeHandle for WryHandle {
       .dispatcher_context
       .proxy
       .send_event(Message::CreateWebview(
-        Arc::new(Mutex::new(Some(Box::new(move |event_loop| {
-          create_webview(event_loop, dispatcher_context, pending)
-        })))),
+        Box::new(move |event_loop| create_webview(event_loop, dispatcher_context, pending)),
         tx,
       ))
       .map_err(|_| Error::FailedToSendMessage)?;
@@ -1266,10 +1314,7 @@ impl Runtime for Wry {
     self.clipboard_manager_handle.clone()
   }
 
-  fn create_window<P: Params<Runtime = Self>>(
-    &self,
-    pending: PendingWindow<P>,
-  ) -> Result<DetachedWindow<P>> {
+  fn create_window(&self, pending: PendingWindow<Self>) -> Result<DetachedWindow<Self>> {
     let label = pending.label.clone();
     let proxy = self.event_loop.create_proxy();
     let webview = create_webview(
@@ -1307,7 +1352,7 @@ impl Runtime for Wry {
   }
 
   #[cfg(feature = "system-tray")]
-  fn system_tray<I: MenuId>(&self, system_tray: SystemTray<I>) -> Result<Self::TrayHandler> {
+  fn system_tray(&self, system_tray: SystemTray) -> Result<Self::TrayHandler> {
     let icon = system_tray
       .icon
       .expect("tray icon not set")
@@ -1373,14 +1418,14 @@ impl Runtime for Wry {
           EventLoopIterationContext {
             callback: &callback,
             webviews: webviews.lock().expect("poisoned webview collection"),
-            window_event_listeners: window_event_listeners.clone(),
+            window_event_listeners: &window_event_listeners,
             global_shortcut_manager: global_shortcut_manager.clone(),
-            global_shortcut_manager_handle: global_shortcut_manager_handle.clone(),
+            global_shortcut_manager_handle: &global_shortcut_manager_handle,
             clipboard_manager: clipboard_manager.clone(),
             #[cfg(feature = "menu")]
-            menu_event_listeners: menu_event_listeners.clone(),
+            menu_event_listeners: &menu_event_listeners,
             #[cfg(feature = "system-tray")]
-            tray_context: tray_context.clone(),
+            tray_context: &tray_context,
           },
         );
       });
@@ -1409,14 +1454,14 @@ impl Runtime for Wry {
         EventLoopIterationContext {
           callback: &callback,
           webviews: webviews.lock().expect("poisoned webview collection"),
-          window_event_listeners: window_event_listeners.clone(),
+          window_event_listeners: &window_event_listeners,
           global_shortcut_manager: global_shortcut_manager.clone(),
-          global_shortcut_manager_handle: global_shortcut_manager_handle.clone(),
+          global_shortcut_manager_handle: &global_shortcut_manager_handle,
           clipboard_manager: clipboard_manager.clone(),
           #[cfg(feature = "menu")]
-          menu_event_listeners: menu_event_listeners.clone(),
+          menu_event_listeners: &menu_event_listeners,
           #[cfg(feature = "system-tray")]
-          tray_context: tray_context.clone(),
+          tray_context: &tray_context,
         },
       );
     })
@@ -1426,14 +1471,14 @@ impl Runtime for Wry {
 struct EventLoopIterationContext<'a> {
   callback: &'a (dyn Fn(RunEvent) + 'static),
   webviews: MutexGuard<'a, HashMap<WindowId, WebviewWrapper>>,
-  window_event_listeners: WindowEventListeners,
+  window_event_listeners: &'a WindowEventListeners,
   global_shortcut_manager: Arc<Mutex<WryShortcutManager>>,
-  global_shortcut_manager_handle: GlobalShortcutManagerHandle,
+  global_shortcut_manager_handle: &'a GlobalShortcutManagerHandle,
   clipboard_manager: Arc<Mutex<Clipboard>>,
   #[cfg(feature = "menu")]
-  menu_event_listeners: MenuEventListeners,
+  menu_event_listeners: &'a MenuEventListeners,
   #[cfg(feature = "system-tray")]
-  tray_context: TrayContext,
+  tray_context: &'a TrayContext,
 }
 
 fn handle_event_loop(
@@ -1454,6 +1499,11 @@ fn handle_event_loop(
     #[cfg(feature = "system-tray")]
     tray_context,
   } = context;
+  if *control_flow == ControlFlow::Exit {
+    return RunIteration {
+      webview_count: webviews.len(),
+    };
+  }
   *control_flow = ControlFlow::Wait;
 
   match event {
@@ -1469,6 +1519,7 @@ fn handle_event_loop(
       window_id,
       menu_id,
       origin: MenuType::MenuBar,
+      ..
     } => {
       let window_id = window_id.unwrap(); // always Some on MenuBar event
       let event = MenuEvent {
@@ -1485,6 +1536,7 @@ fn handle_event_loop(
       window_id: _,
       menu_id,
       origin: MenuType::ContextMenu,
+      ..
     } => {
       let event = SystemTrayEvent::MenuItemClick(menu_id.0);
       for handler in tray_context.listeners.lock().unwrap().values() {
@@ -1496,21 +1548,25 @@ fn handle_event_loop(
       bounds,
       event,
       position: _cursor_position,
+      ..
     } => {
       let (position, size) = (
         PhysicalPositionWrapper(bounds.position).into(),
         PhysicalSizeWrapper(bounds.size).into(),
       );
       let event = match event {
-        TrayEvent::LeftClick => SystemTrayEvent::LeftClick { position, size },
         TrayEvent::RightClick => SystemTrayEvent::RightClick { position, size },
         TrayEvent::DoubleClick => SystemTrayEvent::DoubleClick { position, size },
+        // default to left click
+        _ => SystemTrayEvent::LeftClick { position, size },
       };
       for handler in tray_context.listeners.lock().unwrap().values() {
         handler(&event);
       }
     }
-    Event::WindowEvent { event, window_id } => {
+    Event::WindowEvent {
+      event, window_id, ..
+    } => {
       if let Some(event) = WindowEventWrapper::from(&event).0 {
         for handler in window_event_listeners
           .lock()
@@ -1526,14 +1582,26 @@ fn handle_event_loop(
       }
       match event {
         WryWindowEvent::CloseRequested => {
-          on_window_close(
-            callback,
-            window_id,
-            &mut webviews,
-            control_flow,
-            #[cfg(feature = "menu")]
-            menu_event_listeners.clone(),
-          );
+          let (tx, rx) = channel();
+          if let Some(w) = webviews.get(&window_id) {
+            callback(RunEvent::CloseRequested {
+              label: w.label.clone(),
+              signal_tx: tx,
+            });
+            if let Ok(true) = rx.try_recv() {
+            } else {
+              on_window_close(
+                callback,
+                window_id,
+                &mut webviews,
+                control_flow,
+                #[cfg(target_os = "linux")]
+                window_event_listeners,
+                #[cfg(feature = "menu")]
+                menu_event_listeners.clone(),
+              );
+            }
+          }
         }
         WryWindowEvent::Resized(_) => {
           if let Err(e) = webviews[&window_id].inner.resize() {
@@ -1544,11 +1612,7 @@ fn handle_event_loop(
       }
     }
     Event::UserEvent(message) => match message {
-      Message::Task(task) => {
-        if let Some(task) = task.lock().unwrap().take() {
-          task();
-        }
-      }
+      Message::Task(task) => task(),
       Message::Window(id, window_message) => {
         if let Some(webview) = webviews.get_mut(&id) {
           let window = webview.inner.window();
@@ -1592,9 +1656,16 @@ fn handle_event_loop(
               tx.send(window.available_monitors().collect()).unwrap()
             }
             #[cfg(windows)]
-            WindowMessage::Hwnd(tx) => {
-              use wry::application::platform::windows::WindowExtWindows;
-              tx.send(Hwnd(window.hwnd() as HWND)).unwrap()
+            WindowMessage::Hwnd(tx) => tx.send(Hwnd(window.hwnd() as HWND)).unwrap(),
+            #[cfg(any(
+              target_os = "linux",
+              target_os = "dragonfly",
+              target_os = "freebsd",
+              target_os = "netbsd",
+              target_os = "openbsd"
+            ))]
+            WindowMessage::GtkWindow(tx) => {
+              tx.send(GtkWindow(window.gtk_window().clone())).unwrap()
             }
             // Setters
             WindowMessage::Center(tx) => {
@@ -1622,22 +1693,13 @@ fn handle_event_loop(
             WindowMessage::Show => window.set_visible(true),
             WindowMessage::Hide => window.set_visible(false),
             WindowMessage::Close => {
-              for handler in window_event_listeners
-                .lock()
-                .unwrap()
-                .get(&window.id())
-                .unwrap()
-                .lock()
-                .unwrap()
-                .values()
-              {
-                handler(&WindowEvent::CloseRequested);
-              }
               on_window_close(
                 callback,
                 id,
                 &mut webviews,
                 control_flow,
+                #[cfg(target_os = "linux")]
+                window_event_listeners,
                 #[cfg(feature = "menu")]
                 menu_event_listeners.clone(),
               );
@@ -1669,8 +1731,9 @@ fn handle_event_loop(
             WindowMessage::SetIcon(icon) => {
               window.set_window_icon(Some(icon));
             }
-            WindowMessage::SetSkipTaskbar(skip) => {
-              window.set_skip_taskbar(skip);
+            WindowMessage::SetSkipTaskbar(_skip) => {
+              #[cfg(any(target_os = "windows", target_os = "linux"))]
+              window.set_skip_taskbar(_skip);
             }
             WindowMessage::DragWindow => {
               let _ = window.drag_window();
@@ -1708,22 +1771,16 @@ fn handle_event_loop(
           }
         }
       }
-      Message::CreateWebview(handler, sender) => {
-        let handler = {
-          let mut lock = handler.lock().expect("poisoned create webview handler");
-          std::mem::take(&mut *lock).unwrap()
-        };
-        match handler(event_loop) {
-          Ok(webview) => {
-            let window_id = webview.inner.window().id();
-            webviews.insert(window_id, webview);
-            sender.send(window_id).unwrap();
-          }
-          Err(e) => {
-            eprintln!("{}", e);
-          }
+      Message::CreateWebview(handler, sender) => match handler(event_loop) {
+        Ok(webview) => {
+          let window_id = webview.inner.window().id();
+          webviews.insert(window_id, webview);
+          sender.send(window_id).unwrap();
         }
-      }
+        Err(e) => {
+          eprintln!("{}", e);
+        }
+      },
       #[cfg(feature = "system-tray")]
       Message::Tray(tray_message) => match tray_message {
         TrayMessage::UpdateItem(menu_id, update) => {
@@ -1813,6 +1870,7 @@ fn on_window_close<'a>(
   window_id: WindowId,
   webviews: &mut MutexGuard<'a, HashMap<WindowId, WebviewWrapper>>,
   control_flow: &mut ControlFlow,
+  #[cfg(target_os = "linux")] window_event_listeners: &WindowEventListeners,
   #[cfg(feature = "menu")] menu_event_listeners: MenuEventListeners,
 ) {
   if let Some(webview) = webviews.remove(&window_id) {
@@ -1824,6 +1882,21 @@ fn on_window_close<'a>(
     *control_flow = ControlFlow::Exit;
     callback(RunEvent::Exit);
   }
+  // TODO: tao does not fire the destroyed event properly
+  #[cfg(target_os = "linux")]
+  {
+    for handler in window_event_listeners
+      .lock()
+      .unwrap()
+      .get(&window_id)
+      .unwrap()
+      .lock()
+      .unwrap()
+      .values()
+    {
+      handler(&WindowEvent::Destroyed);
+    }
+  }
 }
 
 fn center_window(window: &Window) -> Result<()> {
@@ -1839,10 +1912,10 @@ fn center_window(window: &Window) -> Result<()> {
   }
 }
 
-fn create_webview<P: Params<Runtime = Wry>>(
+fn create_webview(
   event_loop: &EventLoopWindowTarget<Message>,
   context: DispatcherContext,
-  pending: PendingWindow<P>,
+  pending: PendingWindow<Wry>,
 ) -> Result<WebviewWrapper> {
   #[allow(unused_mut)]
   let PendingWindow {
@@ -1918,7 +1991,7 @@ fn create_webview<P: Params<Runtime = Wry>>(
     .map_err(|e| Error::CreateWebview(Box::new(e)))?;
 
   Ok(WebviewWrapper {
-    label: format!("{}", label),
+    label,
     inner: webview,
     #[cfg(feature = "menu")]
     menu_items,
@@ -1928,10 +2001,10 @@ fn create_webview<P: Params<Runtime = Wry>>(
 }
 
 /// Create a wry rpc handler from a tauri rpc handler.
-fn create_rpc_handler<P: Params<Runtime = Wry>>(
+fn create_rpc_handler(
   context: DispatcherContext,
-  label: P::Label,
-  handler: WebviewRpcHandler<P>,
+  label: String,
+  handler: WebviewRpcHandler<Wry>,
 ) -> Box<dyn Fn(&Window, WryRpcRequest) -> Option<RpcResponse> + 'static> {
   Box::new(move |window, request| {
     handler(
@@ -1949,10 +2022,10 @@ fn create_rpc_handler<P: Params<Runtime = Wry>>(
 }
 
 /// Create a wry file drop handler from a tauri file drop handler.
-fn create_file_drop_handler<P: Params<Runtime = Wry>>(
+fn create_file_drop_handler(
   context: DispatcherContext,
-  label: P::Label,
-  handler: FileDropHandler<P>,
+  label: String,
+  handler: FileDropHandler<Wry>,
 ) -> Box<dyn Fn(&Window, WryFileDropEvent) -> bool + 'static> {
   Box::new(move |window, event| {
     handler(
